

#target tap

; tape block sync bytes:
headerflag   	equ 	0
dataflag     	equ 	0xff

; some Basic tokens:
tCLEAR		equ     $FD             ; token CLEAR
tLOAD		equ     $EF             ; token LOAD
tCODE		equ     $AF             ; token CODE
tPRINT		equ     $F5             ; token PRINT
tRANDOMIZE	equ     $F9             ; token RANDOMIZE
tUSR		equ     $C0             ; token USR

; physical memory layout:
pixels_start	equ	0x4000		; ZXSP screen pixels
attr_start	equ	0x5800		; ZXSP screen attributes
attr_end	equ	0x5B00		; ZXSP printer buffer
slow_ram	equ 	0x5B00		; start of usable slow ram
fast_ram	equ 	0x8000		; start of fast ram
ram_end 	equ 	$10000

; memory usage:
load_address	equ	$C000		; code block is initially loaded here
code_address	equ	attr_end	; will be moved here
code_size	equ	code_size	; size of code block on tape (calculated at end of file)
start_address	equ	decompress	; mcode start address

stack_size	equ 	512		; VM return + VM variables total stack size
rstack_size	equ	40 * 2		; max. nesting proc level



; ---------------------------------------------------
;	a Basic Loader:
; ---------------------------------------------------

#code PROG_HEADER,0,17,headerflag
	defb    0			; Indicates a Basic program
	defb    "mloader   "		; the block name, 10 bytes long
	defw    variables_end-0		; length of block = length of basic program plus variables
	defw    10		    	; line number for auto-start, 0x8000 if none
	defw    program_end-0		; length of the basic program without variables

#code PROG_DATA,0,*,dataflag

; ZX Spectrum Basic:

; 10 CLEAR load_address-1
        db  0,10		; line number
        db  end10-($+1)		; line length
        db  0			; statement number
        db  tCLEAR		; token CLEAR
        db  "0", $0e0000 \ dw load_address-1 \ db 0   ; number: ascii & internal format
end10:	db  $0d			; line end marker

; 20 LOAD "" CODE
        db  0,20		; line number
        db  end20-($+1)		; line length
        db  0			; statement number
        db  tLOAD, '""', tCODE	; token LOAD, 2 quotes, token CODE
end20:	db  $0d			; line end marker

; 30 RANDOMIZE USR start_address
        db  0,30		; line number
        db  end30-($+1)		; line length
        db  0			; statement number
        db  tRANDOMIZE, tUSR	; token RANDOMIZE, token USR
        db  "0", $0e0000 \ dw start_address \ db 0   ; number: ascii & internal format
end30:	db  $0d			; line end marker

program_end:

	; ZX Spectrum Basic variables

variables_end:



; ---------------------------------------------------
;	The machine code block:
; ---------------------------------------------------

#code CODE_HEADER, 0, 17, headerflag
	defb    3			; Indicates binary data
	defb    "mcode     "	  	; the block name, 10 bytes long
	defw    code_size		; length of data block which follows
	defw    load_address		; default loading location for the data
	defw    0       		; unused

#code CODE_DATA, 0, 0, dataflag

; Machine Code and Data:
; note: all following #code blocks are appended to #code CODE_DATA



; ---------------------------------------------------
;	MEMORY LAYOUT
; ---------------------------------------------------

; we put everything in #code segments
; this is very handy with zx7 compression

; slow ram:
#code BOARDS, code_address	; compressed boards
#code LEVEL1			; compressible TODO
#code LEVEL2			; compressible TODO
#code LEVEL3			; compressible TODO
#code LEVEL4			; compressible TODO
#code SLOW_CDATA		; compressed or uncompressible data, e.g. compressed levels, music
uncompressed_data_size:: equ SLOW_CDATA_end - code_address
#code SLOW_DATA			; Uncritical/interrupt ram: variables and mutable code
#code SLOW_CODE 		; Uncritical/interrupt rom: immutable romable code

; fast ram:
#code VCODE			; rom: All virtual code (except boot code): Main Program, Libraries
#code VMCORE			; rom: VM core
#code CODE			; rom: Immutable romable machine code 
#code CONST			; rom: Const data
#code GSDATA					; initialized data: variables and mutable code, Vcode GVARS
#code DATA					; ram: Variables and mutable code, Vcode GSVARS
#data STACKS, (DATA_end+1)&$FFFE, stack_size 	; ram: VM return stack + VM variables stack
#data HEAP, STACKS_end, ram_end-$$		; ram: Managed memory


; boot code:
#code GSINIT, HEAP		 ; rom: boot code: mcode + vcode -> compressed, relocated, inside heap!
#code ZX7, load_address + uncompressed_data_size + SLOW_DATA_to_DATA_csize + GSINIT_csize  
				; rom: uncompressed, unrelocated code

; compress everything except uncompressed_data_size and ZX7:
#compress SLOW_DATA to DATA	
#compress GSINIT
#assert load_address + uncompressed_data_size + SLOW_DATA_to_DATA_csize >= DATA_end + SLOW_DATA_to_DATA_cdelta
#assert ZX7 >= GSINIT_end + GSINIT_cdelta




; ---------------------------------------------------
;	DEFINE VM OPTIONS (or NON-options)
; ---------------------------------------------------

; non-options:
BC_IS_VPC		equ 1	; VM program counter
SP_IS_VSTACK		equ 1	; VM variables stack
DE_IS_TOP		equ 1	; TOP value on VSTACK
VSTACK_IS_ALIGNED	equ 1	; VSTACK is word-aligned: no single byte pushing!
IX_IS_RSTACK		equ 1	; VM return stack
; options:
OFFS_IX			equ rstack_size/2
IY_IS_NEXT		equ 1	; IY register always points to next ('jp iy')
HEAP_RETAIN_RELEASE	equ 1	; retain-release model: memory is compacted by garbage collection
HEAP_USES_HANDLES	equ 1	; simplifies garbage collection

dflt_heap equ GSINIT_end	; base address of heap during GSINIT

#code GSDATA 			; global data must all be in one page
	.align	hi(DATA_end) != hi(GSDATA) ? 256 : 1
GVARS	equ $ & 0xff00



; ---------------------------------------------------
;	INTERRUPT HANDLER
; ---------------------------------------------------

#code SLOW_CODE
interrupt_handler:: 
	push	af
	push	bc
	push	de
	push	hl
	call	irpt_scan_keyboard	; lib-zxsp/keyboard.s
	call 	zxsp_kempston_mouse	; lib-zxsp/kempston_mouse.s
	pop	hl
	pop	de
	pop	bc
	pop	af
	ei
	ret



; ---------------------------------------------------
;	SYSTEM VARIABLES
; ---------------------------------------------------

.area DATA
errno::		defs	1
scratch::	defs	10		; print needs 10 byte	



; ---------------------------------------------------
;	BOOT CODE
; ---------------------------------------------------

#code ZX7
decompress:
	di
	ld	sp, STACKS_end
        ld      hl, load_address	        ; source
        ld      de, code_address	        ; dest
#assert uncompressed_data_size != 0
	ld	bc, uncompressed_data_size
	ldir
	call	decompress_zx7_standard

	ld	hl, load_address + uncompressed_data_size + SLOW_DATA_to_DATA_csize
	ld	de, GSINIT
	ld	bc, start \ push bc		; preset return address for ZX7
	;call	decompress_zx7_standard
	;jp	start
#include "lib-z80/decompress_zx7_standard.s"    ; include decompressor


; ________________________________
#code GSINIT

	dw	GSINIT_size - 2			; construct a valid heap block

start:	;di
	;ld	sp, STACKS_end			; VStack
	ld	ix, STACKS + rstack_size/2	; RStack
	
	; setup interrupt vector:
	ld	a, hi(intv)
	ld	i,a
	im	2			

	; clear heap to zero:
	xor 	a
	ld	hl, GSINIT_end
1$:	ld	(hl),a
	inc l \ jp nz,1$		; assumed: ram_end = $10000
	inc h \ jp nz,1$



; --- setup interrupt ---
; adjust position of this code so that the gap before 'intv' is small

#code SLOW_DATA
	ds 	255 - lo($)		; low byte of address of interrupt_vector must be 255
intv::	dw 	interrupt_handler	; interrupt vector



; ===========================================
; 	    INCLUDE EVERYTHING
; ===========================================

#include "lib-zxsp/keyboard.s"
#include "lib-zxsp/kempston_mouse.s"
#include "lib-zxsp/charset_p8.s"
#include "lib-zxsp/screen.s"
#include "lib-zxsp/print_p8.s"
#include "lib-zxsp/draw.s"
#include "lib-vcc/core-macros.s"
#include "lib-vcc/core.s"
#include "lib-vcc/core-memory.s"
#include "system.s"
#include "abort.s"
#include "board.s"
#include "boards.s"
#include "graphics.s"
#code CODE
#include "lib-z80/itoa.s"


; now switch to virtual mode for initializer code generated by Vcc:
#code GSINIT
	virtual
	
#cpath 	 "vcc"	
#cflags  -cpu=vss -DDEBUG -o4 -v -c
#include "main.cc"

#code GSINIT
	dw	IVAL, HEAP		; store address of HEAP start
	dw	GSET			; in gvar heap_start
	db 	heap_start - GVARS	; -> GSINIT block becomes visible as unused heap block
	dw	ei
	dw	JP, DOIT		; jump out of GSINIT to persistent code
	
#code VCODE
DOIT:	dw	main			; call main()
	PANIC 	ERROR_END_OF_CODE	; must not return


; resolve missing labels:
#code VMCORE
#include library "lib-vcc"


; align DATA_end so that preliminary sizes settle
#code DATA
	defs	0
	.align 	4


; TEST:
#code SLOW_DATA
sintab::
angle = 0	
	.rept 256
	 db	sin(angle,256,95)
angle = angle+1
	.endm



; ===========================================
;	FINAL CALCULATIONS
; ===========================================

; size of code block on tape:
code_size	equ	uncompressed_data_size + SLOW_DATA_to_DATA_csize + GSINIT_csize + ZX7_size

; calculated size of tape file:
BLK1SZ		equ 	4 + PROG_HEADER_size
BLK2SZ		equ 	4 + PROG_DATA_size
BLK3SZ		equ 	4 + CODE_HEADER_size
BLK4SZ		equ 	4 + code_size
tape_size	equ	BLK1SZ + BLK2SZ + BLK3SZ + BLK4SZ




